"""List some information about all postprocessing filters available in the directory this module resides in.

This is a development-time introspection tool that is helpful for getting an overview
of the default stageName and sort values of all filters at a glance. This is especially useful
when looking for the appropriate position to place a new type of filter in.

The reported information is module name, class name, and stageName and sort of each Filter found.

This tool should be run in direct/src/filter with

  python -m ListAllFilters

"""

from __future__ import print_function

import imp, os, re
import importlib
import inspect
import collections

# These must be imported first before importing any filters to avoid dependency errors (FIXME).
#
import FilterPipeline
import FilterManager

import Filter
import CompoundFilter

# This is needed to instantiate FilterPipeline (which we need to do to get a list of the default knownStages)
from direct.directbase import DirectStart

# The module finder comes from a tool in Python 2.5, by Fredrik Lundh, 2005:
#   http://svn.python.org/projects/python/tags/r252/Doc/tools/listmodules.py

_suffixes = imp.get_suffixes()
def _get_suffix(file):
    for suffix in _suffixes:
        if file[-len(suffix[0]):] == suffix[0]:
            return suffix
    return None

def getmodules(p):
    """Get modules in a given directory p (string, path)."""
    modules = {}
    for f in os.listdir(p):
        f = os.path.join(p, f)
        if os.path.isfile(f):
            m, e = os.path.splitext(f)
            suffix = _get_suffix(f)
            if not suffix:
                continue
            m = os.path.basename(m)
            if re.compile("(?i)[a-z_]\w*$").match(m):
                if suffix[2] == imp.C_EXTENSION:
                    # check that this extension can be imported
                    try:
                        __import__(m)
                    except ImportError:
                        continue
                modules[m] = f
        elif os.path.isdir(f):
            m = os.path.basename(f)
            if os.path.isfile(os.path.join(f, "__init__.py")):
                for mm, f in getmodules(f).items():
                    modules[m + "." + mm] = f
    return modules


# Our strategy is to find all modules, import each of them,
# and find all subclasses of Filter defined in each.

# We blacklist some modules which do not define any filters themselves, but import some
# (either directly or via other imports). Importing these modules would confuse the inspection logic.
#
moduleBlacklist = ["CommonFilters", "FilterPipeline"]

RawFilterInfo = collections.namedtuple( "RawFilterInfo", ["modName", "clsName", "cls"] )

def getFilters(p):
    """Get Filters defined in a given directory p (string, path)."""
    filterModules = getmodules(p)
    rawFilterInfos = []
    for modName in filterModules.keys():
        if modName in moduleBlacklist:
            continue
        mod = importlib.import_module(modName)

        thisModRawFilterInfos = []
        for clsName in dir(mod):
            cls = getattr(mod, clsName)
            if not inspect.isclass(cls):
                continue
            if issubclass(cls, Filter.Filter):
                thisModRawFilterInfos.append( RawFilterInfo(modName, clsName, cls) )

        # If the module defines a compound filter, any other filters that module "defines"
        # are likely duplicates caused by importing the contained filters in that module.
        # We would like to ignore those.
        #
        # TODO/FIXME: This will break if a compound filter is added to MiscFilters (which defines many filters).
        # TODO/FIXME: We need a better way to detect which names were caused by imports.
        #
        temp = map( lambda item: issubclass(item.cls, CompoundFilter.CompoundFilter), thisModRawFilterInfos )
        isAnyCompound = len(temp) > 0  and  max(temp) == True
        if isAnyCompound:
            # This module defines at least one compound filter.
            # Add only the compound filter(s) to the filter infos.
            #
            for item in thisModRawFilterInfos:
                if issubclass(item.cls, CompoundFilter.CompoundFilter):
                    rawFilterInfos.append( item )
        else:
            # This module did not define any compound filter. All filters defined by the module
            # should be valid (not import-caused duplicates). Add all of them to the filter infos.
            #
            rawFilterInfos.extend( thisModRawFilterInfos )

    return rawFilterInfos


# The filter search will skip trying to instantiate classes with these names.
#
# TODO/FIXME: We use this brittle solution because there is no reliable way to dynamically detect
# TODO/FIXME: which of the filter classes are abstract.
#
knownAbstractClassNames = ["Filter", "CompoundFilter", "RapidPrototypeCompoundFilter"]

PrintableFilterInfo = collections.namedtuple( "PrintableFilterInfo", ["modName", "clsName", "stageName", "sort"] )

def report(path=".", sortKeyName="render"):
    """Report Filters in the directory "path" into stdout.

    The report is printed in a fixed-width format.

    The ordering is controlled by sortKeyName:
      "render"      : order by stageName (in logical order), then the sort value
      "class"       : order by just the class name, globally
      "moduleclass" : order by module name, then by class name within each module

    """

    filterInfos = []
    for rawFilterInfo in getFilters(p=path):
        if rawFilterInfo.clsName in knownAbstractClassNames:
            continue
        instance = rawFilterInfo.cls()
        filterInfos.append( PrintableFilterInfo(rawFilterInfo.modName, rawFilterInfo.clsName,
                                                instance.stageName, instance.sort) )
    headerItem = PrintableFilterInfo("Module", "Class", "stageName", "sort")

    # Instantiate a FilterPipeline to extract the default knownStages.
    #
    # For this we need DirectStart, because FilterPipeline references the global taskMgr,
    # and initializing FilterManager (inside FilterPipeline) requires a window and a camera.
    #
    fpp = FilterPipeline.FilterPipeline(win=base.win, cam=base.cam)
    k = [None] + fpp.knownStages  # StageInitializer has the special value stageName=None,
                                  # which is not valid for any other filter, so it is not
                                  # present in knownStages.

    sortKeys = { "render"      : lambda item: (k.index(item.stageName), item.sort),
                 "class"       : lambda item: item.clsName,
                 "moduleclass" : lambda item: (item.modName, item.clsName) }

    filterInfos.sort( key=sortKeys[sortKeyName] )
    filterInfos = [headerItem] + filterInfos

    # Perform some fixed-width formatting in the output.
    #
    def columnLength(col, item):
        return len( str(item[col]) )

    maxColumnLengths = []
    for col in xrange(len(headerItem)):  # each item is assumed to have the same number of columns as the header.
        maxColumnLengths.append(  max( map( lambda item : columnLength(col, item), filterInfos ) )  )

    prevStageName = None
    for lineno,item in enumerate(filterInfos):
        isHeaderLine = (lineno == 0)

        columns = []
        for col,text in enumerate(item):
            pad = " " * (maxColumnLengths[col] - columnLength(col, item))
            columns.append( "%s%s" % (text, pad) )
        line = " ".join(columns)  # observe that line has a constant width.

        # When reporting in render order, mark logical stage boundaries
        # (special-casing the header and the initial "None").
        #
        if sortKeyName == "render"  and  not isHeaderLine  and  item.stageName is not None:
            if item.stageName != prevStageName:
                print( "-" * len(line) )
                prevStageName = item.stageName

        print( line )

        # Print separator after header.
        #
        if isHeaderLine:
            print( "=" * len(line) )


def main():
    """Print reports with all three orderings."""

    fullPathOfThisModule = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
    print("\nPostprocessing filter report for directory '%s'\n" % fullPathOfThisModule)

    print("\nAll filters by class name:\n")
    report(path=fullPathOfThisModule, sortKeyName="class")
    print("")

    print("\nAll filters by module, then class:\n")
    report(path=fullPathOfThisModule, sortKeyName="moduleclass")
    print("")

    print("\nAll filters in render order (when using default settings):\n")
    report(path=fullPathOfThisModule, sortKeyName="render")
    print("")


if __name__ == '__main__':
    main()

