"""A collection of short, simple filters for FilterPipeline.

More complex filters each have their own module.

"""

from pandac.PandaModules import AuxBitplaneAttrib, Vec4

from Filter import Filter, TextureInfo

###################################################
# Special magic filter to initialize a render pass
###################################################

class StageInitializer(Filter):
    """This filter simply initializes pixcolor from the current pixel in the pipeline stage input color texture.

    It is automatically inserted into each FilterStage created during FilterPipeline.reconfigure().

    Unlike other filters, the properties of StageInitializer are not meaningful,
    and the filter is owned by (the only is reference stored in) the FilterStage it resides in.
    Hence, during each reconfigure of the pipeline, old StageInitializers are destroyed,
    and new ones created (because FilterStages are).

    The sort value must be -1, which places StageInitializer before any other filters in the stage.
    Only StageInitializer is allowed to have a sort value of -1.

    """
    def __init__(self, **kwargs):
        super(StageInitializer, self).__init__(**kwargs)

        # Unlike usual, this filter doesn't want the compositing fshader
        # to supply the current value of pixcolor to the filter function
        # (because the purpose of StageInitializer is to
        #  determine and assign the initial value of pixcolor).
        #
        # For more information, see Filter.__init__().
        #
        self._needPixcolor = False

    def onReset(self):
        super(StageInitializer, self).onReset()  # reset inherited properties
        self.sort        = -1     # Initialization goes first before any other filter
                                  # in the same FilterStage.

        self.stageName   = None   # This filter is special; the value of stageName is not used.
                                  # Instead, each created render pass gets one copy of the filter.

        self.isMergeable = None   # The initialization filter is only added after the logical stage
                                  # into render pass merging process is complete. The value of isMergeable
                                  # is ignored.

    def onAttachPipeline(self):
        # The compositing shader needs the color texture to initialize the output color.
        self.registerInputTexture(texName="color")

    def onSynthesizeCompositor(self):
        # Registered textures are accessed with self.getTextureInfo(texName).
        # This ensures that the correct definition is returned.
        #
        # See Filter.TextureInfo for the available fields.
        #
        # pixcolor is the input/output variable for the color of the current pixel.
        #
        txcolor = self.getTextureInfo("color")
        return ("stageInitializer",
                "float4 pixcolor = tex2D(%(k_txcolor)s, %(texcoord_txcolor)s);\n" % { "k_txcolor" : txcolor.varname,
                                                                                      "texcoord_txcolor" : txcolor.texcoord },
                "// initialize pixcolor\n")


###################################################
# Actual regular MiscFilters
###################################################

class Tint(Filter):
    """Color tint / fade-to-color filter.

    Introduced in Panda 1.9.0.

    In "multiply" mode [default]:

      The color of each pixel is multiplied by the specified constant color.
      The result is saturated to [0,1] in each component.

      The tinted pixel then is blended with the original one with the strength
      specified by "strength".

    In "add" mode:

      The specified constant color is added to the color of each pixel.

      Note that it is valid for one or more of the components of the constant color
      to be negative; negative components are effectively subtracted.

      The result is saturated to [0,1] in each component.

      The tinted pixel then is blended with the original one with the strength
      specified by "strength".

    In "fade" mode:

      Each pixel is blended toward the specified constant color.
      The parameter "strength" controls the strength.

    This filter can be used e.g. to fade the screen to black (color=(0,0,0,1))
    by dynamically updating the "strength" parameter in your own update task
    so that it gradually goes from 0 to 1. (Make your task sort 48 or earlier
    to ensure it runs before filter reconfigure/update and frame redraw (igLoop).)

    The fade mode also allows fading toward colors other than black
    (e.g. color=(1,1,1,1) to fade to white).

    In tint mode, fading the color to (255,255,255,1) gives another kind of
    fade to white (using the fact that there are 8 bits per color channel).

    Parameters:

      mode  = string, one of:
                "multiply" = tint (multiply) by the specified color.
                "add"      = add the specified color.
                "fade"     = fade to the specified color.

      color = (R,G,B,A) tuple; tint color. How exactly this is used depends on "mode"; see above.

              Usually setting the alpha component to 1 is a good idea in
              "multiply" and "fade" modes; in "add" mode, alpha should usually be 0.

      strenth = strength of blending between the original and processed color.
                1.0 means fully processed, 0.0 does nothing.
                Behaves linearly.

    """
    def __init__(self, **kwargs):
        super(Tint, self).__init__(**kwargs)

    def onReset(self):
        super(Tint, self).onReset()  # reset inherited properties
        self.sort  = 90  # late within the pipeline stage
        self.stageName = "Postprocess"
        self.color = (0.5,0.5,0.5,1)  # default is darken by 50%
        self.strength = 1.0
        self.mode  = "multiply"  # compile-time parameter

    def onAttachPipeline(self):
        # We do not need to register the "color" texture, as our code does not access it explicitly,
        # and we do not need texture coordinates, either.
        #
        # The StageInitializer sets up the initial value of pixcolor automatically.
        # It is thus better to refer to pixcolor for the original color of the pixel being processed,
        # to save one texture lookup (and unnecessary parameters to our Cg function generated
        # in synthesize()).
        #
        self.registerCustomInput(inputType="float4", inputName="k_tint_color")
        self.registerCustomInput(inputType="float",  inputName="k_tint_strength")

    def onSynthesizeCompositor(self):
        code = ""
        if self.mode == "multiply":
            code += "// Compile-time mode: multiply\n"
            code += "float4 tinted_color = saturate(pixcolor * %(k_tint_color)s);\n" % { "k_tint_color" : self.getMangledName("k_tint_color") }
            code += "pixcolor = lerp(pixcolor, tinted_color, %(k_tint_strength)s);\n" % { "k_tint_strength" : self.getMangledName("k_tint_strength") }
        elif self.mode == "add":
            code += "// Compile-time mode: add\n"
            code += "float4 tinted_color = saturate(pixcolor + %(k_tint_color)s);\n" % { "k_tint_color" : self.getMangledName("k_tint_color") }
            code += "pixcolor = lerp(pixcolor, tinted_color, %(k_tint_strength)s);\n" % { "k_tint_strength" : self.getMangledName("k_tint_strength") }
        elif self.mode == "fade":
            code += "// Compile-time mode: fade\n"
            code += "pixcolor = lerp(pixcolor, %(k_tint_color)s, %(k_tint_strength)s);\n" % \
                        { "k_tint_color" : self.getMangledName("k_tint_color"),
                          "k_tint_strength" : self.getMangledName("k_tint_strength") }
        else:
            raise ValueError("Tint: invalid mode '%s'; valid: 'multiply', 'add', 'fade'" % self.mode)
        return ("tint", code, "// Color tint / fade-to-color filter\n")


    @property
    def color(self):
        """Tint color as (R,G,B,A) float tuple, each component in [0,1]."""
        return self._color
    @color.setter
    def color(self, value):
        self._color = value
        if self.finalQuad is not None:
            self.finalQuad.setShaderInput( self.getMangledName("tint_color"), value )

    @property
    def strength(self):
        """Effect strength, float. 1 is full strength, 0 is off."""
        return self._strength
    @strength.setter
    def strength(self, value):
        self._strength = value
        if self.finalQuad is not None:
            self.finalQuad.setShaderInput( self.getMangledName("tint_strength"), value )

    @property
    def mode(self):
        """Effect mode, one of 'multiply' (default), 'add' or 'fade'."""
        return self._mode
    @mode.setter
    def mode(self, value):
        if (not hasattr(self, "_mode")  or  value != self._mode)  and  self.pipeline is not None:
            self.pipeline._needsCompile = True
        self._mode = value


class ViewGlow(Filter):
    """View the glow map by overwriting the red color channel with the glow data. This may help in debugging your scene."""

    def __init__(self, **kwargs):
        super(ViewGlow, self).__init__(**kwargs)

    def onReset(self):
        super(ViewGlow, self).onReset()
        self.stageName = "DebugHelpers"  # this is a debug helper filter which should go last
        self.sort = 99  # last thing in the stage

    def onAttachPipeline(self):
        self.requireAuxBits(bitmask=AuxBitplaneAttrib.ABOGlow)

    def onSynthesizeCompositor(self):
        return ("viewGlow",
                "pixcolor.r = pixcolor.a;\n",
                "// show the glow map by overwriting the red channel (debugging helper filter)\n")


class GammaAdjust(Filter):
    """Apply additional gamma correction to the image.

    Introduced in Panda 1.9.0.

    """

    def __init__(self, **kwargs):
        super(GammaAdjust, self).__init__(**kwargs)

    def onReset(self):
        super(GammaAdjust, self).onReset()
        self.stageName = "DisplayDevice"
        self.sort = 90

    def onAttachPipeline(self):
        # This filter needs no textures, custom inputs, aux bits, or internal stages,
        # but this method is abstract in the Filter base class. Thus, we just provide
        # a blank implementation to fulfill the interface and acknowledge that
        # this filter doesn't need to do anything here.
        #
        pass

    def onSynthesizeCompositor(self):
        if self.gamma == 0.5:
            code  = "// gamma = 0.5\n"
            code += "pixcolor.rgb = sqrt(pixcolor.rgb);\n"
        elif self.gamma == 2.0:
            code  = "// gamma = 2.0\n"
            code += "pixcolor.rgb *= pixcolor.rgb;\n"
        elif self.gamma != 1.0:
            code  = "// gamma = %g\n" % self.gamma
            code += "pixcolor.rgb = pow(pixcolor.rgb, %ff);\n" % self.gamma
        else: # self.gamma == 1.0:  # no-op
            code  = "// gamma = 1.0, no-op\n"

        return ("gammaAdjust",
                code,
                "// gamma correction\n")

    @property
    def gamma(self):
        """Amount of gamma correction. 1.0 = no correction."""
        return self._gamma
    @gamma.setter
    def gamma(self, value):
        if (not hasattr(self, "_gamma")  or  value != self._gamma)  and  self.pipeline is not None:
            self.pipeline._needsCompile = True
        self._gamma = value


class ColorInversion(Filter):
    """Invert the colors, like a film negative."""

    def __init__(self, **kwargs):
        super(ColorInversion, self).__init__(**kwargs)

    def onReset(self):
        super(ColorInversion, self).onReset()
        self.stageName = "FilmOrDetector"
        self.sort = 90

    def onAttachPipeline(self):
        pass

    def onSynthesizeCompositor(self):
        return ("colorInversion",
                "pixcolor = float4(1.0, 1.0, 1.0, 1.0) - pixcolor;\n",
                "// invert colors\n")


# Inking algorithm from Panda 1.8.1, with a bugfix: the screen-space normal in the aux texture
# has three components, not two. Changed the weighting float4(3,3,0,0) to float4(2,2,2,0).
#
CARTOONINK_BODY="""float4 cartoondelta = %(k_cartoonseparation)s * %(texpix_txaux)s.xwyw;
float4 cartoon_c0 = tex2D(%(k_txaux)s, %(texcoord_txaux)s + cartoondelta.xy);
float4 cartoon_c1 = tex2D(%(k_txaux)s, %(texcoord_txaux)s - cartoondelta.xy);
float4 cartoon_c2 = tex2D(%(k_txaux)s, %(texcoord_txaux)s + cartoondelta.wz);
float4 cartoon_c3 = tex2D(%(k_txaux)s, %(texcoord_txaux)s - cartoondelta.wz);
float4 cartoon_mx = max(cartoon_c0, max(cartoon_c1, max(cartoon_c2, cartoon_c3)));
float4 cartoon_mn = min(cartoon_c0, min(cartoon_c1, min(cartoon_c2, cartoon_c3)));
float cartoon_thresh = saturate(dot(cartoon_mx - cartoon_mn, float4(2,2,2,0)) - 0.5);
pixcolor = lerp(pixcolor, %(k_cartooncolor)s, cartoon_thresh);
"""

class CartoonInkClassic(Filter):
    """A cartoon outline inking filter.

    The inking is based on examining discontinuities in the normal map, as viewed from the camera.

    Classic algorithm, as in Panda 1.8.1.

    """
    def __init__(self, **kwargs):
        super(CartoonInkClassic, self).__init__(**kwargs)

    def onReset(self):
        super(CartoonInkClassic, self).onReset()
        # Inking must come before pretty much everything else, to simulate a completely drawn cel.
        self.stageName  = "Preprocess"
        self.sort       = 50
        self.separation = 1.0
        self.color      = (0.0, 0.0, 0.0, 1.0)

    def onAttachPipeline(self):
        self.registerInputTexture(texName="aux")
        self.requireAuxBits(bitmask=AuxBitplaneAttrib.ABOAuxNormal)

        self.registerCustomInput(inputType="float",  inputName="k_cartoonseparation")
        self.registerCustomInput(inputType="float4", inputName="k_cartooncolor")

    def onSynthesizeCompositor(self):
        txaux = self.getTextureInfo("aux")
        code = CARTOONINK_BODY % { "k_cartoonseparation" : self.getMangledName("k_cartoonseparation"),
                                   "k_cartooncolor"      : self.getMangledName("k_cartooncolor"),
                                   "k_txaux"             : txaux.varname,
                                   "texcoord_txaux"      : txaux.texcoord,
                                   "texpix_txaux"        : txaux.texpix }

        return ("cartoonInkClassic",
                code,
                "// cartoon ink (outlines), classic algorithm\n")


    @property
    def separation(self):
        """Stencil size for examining discontinuities, in pixels.

        Float, default 1.0.

        Pixels in the +x, -x, +y and -y directions from the current pixel
        will be examined, with the distance for each check set to 'separation'.

        Float. Values in the range 0.6 ... 1.0 are usually good. Some scenes may tolerate
        larger values; try it on your scene to see.

        """
        return self._separation
    @separation.setter
    def separation(self, value):
        self._separation = value
        if self.finalQuad is not None:
            self.finalQuad.setShaderInput( self.getMangledName("cartoonseparation"), value )


    @property
    def color(self):
        """Color to use for ink, as (R,G,B,A) tuple. [introduced in 1.8.0]

        Default (0.0, 0.0, 0.0, 1.0).

        The alpha component gives the _maximum_ alpha value that corresponds
        to a fully inked pixel.

        """
        return self._color
    @color.setter
    def color(self, value):
        self._color = value
        if self.finalQuad is not None:
            self.finalQuad.setShaderInput( self.getMangledName("cartooncolor"), Vec4(value) )


# Algorithm by Timothy Lottes of NVIDIA; implementation based on
# http://horde3d.org/wiki/index.php5?title=Shading_Technique_-_FXAA
#
FXAA_BODY="""float FXAA_SPAN_MAX = 8.0;
float FXAA_REDUCE_MUL = 1.0/8.0;
float FXAA_REDUCE_MIN = 1.0/128.0;

float3 rgbNW = tex2D(%(k_txcolor)s, %(texcoord_txcolor)s + float2(-1.0, -1.0) * %(texpix_txcolor)s).rgb;
float3 rgbNE = tex2D(%(k_txcolor)s, %(texcoord_txcolor)s + float2( 1.0, -1.0) * %(texpix_txcolor)s).rgb;
float3 rgbSW = tex2D(%(k_txcolor)s, %(texcoord_txcolor)s + float2(-1.0,  1.0) * %(texpix_txcolor)s).rgb;
float3 rgbSE = tex2D(%(k_txcolor)s, %(texcoord_txcolor)s + float2( 1.0,  1.0) * %(texpix_txcolor)s).rgb;
float3 rgbM  = pixcolor;  // reuse existing center tap

vec3 luma = vec3(0.299, 0.587, 0.114);
float lumaNW = dot(rgbNW, luma);
float lumaNE = dot(rgbNE, luma);
float lumaSW = dot(rgbSW, luma);
float lumaSE = dot(rgbSE, luma);
float lumaM  = dot(rgbM,  luma);

float lumaMin = min(lumaM, min(min(lumaNW, lumaNE), min(lumaSW, lumaSE)));
float lumaMax = max(lumaM, max(max(lumaNW, lumaNE), max(lumaSW, lumaSE)));

float2 dir;
dir.x = -( (lumaNW + lumaNE) - (lumaSW + lumaSE) );
dir.y =  ( (lumaNW + lumaSW) - (lumaNE + lumaSE) );

float dirReduce = max(
        (lumaNW + lumaNE + lumaSW + lumaSE) * (0.25 * FXAA_REDUCE_MUL),
        FXAA_REDUCE_MIN);

float rcpDirMin = 1.0 / ( min(abs(dir.x), abs(dir.y)) + dirReduce );

dir = min(     float2( FXAA_SPAN_MAX,  FXAA_SPAN_MAX),
           max(float2(-FXAA_SPAN_MAX, -FXAA_SPAN_MAX), dir * rcpDirMin)
         ) * %(texpix_txcolor)s;

vec3 rgbA = (1.0/2.0) * (
        tex2D( %(k_txcolor)s, %(texcoord_txcolor)s + dir * (1.0/3.0 - 0.5) ).rgb +
        tex2D( %(k_txcolor)s, %(texcoord_txcolor)s + dir * (2.0/3.0 - 0.5) ).rgb
      );
vec3 rgbB = rgbA * (1.0/2.0) + (1.0/4.0) * (
        tex2D( %(k_txcolor)s, %(texcoord_txcolor)s + dir * (0.0/3.0 - 0.5) ).rgb +
        tex2D( %(k_txcolor)s, %(texcoord_txcolor)s + dir * (3.0/3.0 - 0.5) ).rgb
      );
float lumaB = dot(rgbB, luma);

if( (lumaB < lumaMin) || (lumaB > lumaMax) )
    pixcolor.rgb = rgbA;
else
    pixcolor.rgb = rgbB;
"""

class AntialiasFXAA(Filter):
    """Fast approximate antialiasing (FXAA).

    This is a fast postprocess full-screen antialiasing filter based on Timothy Lottes's FXAA algorithm.

    This smooths jaggy edges, but also blurs sharp edges and high-frequency components in textures;
    this should be applied before drawing any UI elements.

    Introduced in Panda 1.9.0.

    """
    def __init__(self, **kwargs):
        super(AntialiasFXAA, self).__init__(**kwargs)

    def onReset(self):
        super(AntialiasFXAA, self).onReset()
        self.stageName   = "Preprocess"
        # FXAA must come first in whichever stage it is in, because it overwrites pixcolor
        # (and will thus erase any earlier processing performed in the same stage).
        self.sort        = 0
        self.isMergeable = False

    def onAttachPipeline(self):
        self.registerInputTexture(texName="color")

    def onSynthesizeCompositor(self):
        txcolor = self.getTextureInfo("color")
        code = FXAA_BODY % { "k_txcolor"        : txcolor.varname,
                             "texcoord_txcolor" : txcolor.texcoord,
                             "texpix_txcolor"   : txcolor.texpix }

        return ("antialiasFXAA",
                code,
                "// fast approximate antialiasing\n")


VIGNETTING_BODY="""// TODO/FIXME: is texpad always < 0.5? In that case we don't need the min(), and can use texpad directly.
const float2 texradii = min( %(texpad_txcolor)s.xy, 1.0 - %(texpad_txcolor)s.xy );
const float2 sample_vector = (%(texpad_txcolor)s.xy - %(texcoord_txcolor)s.xy)/texradii.xy; // in -1 ... 1
const float  sample_normsq = dot(sample_vector, sample_vector);  // in 0...2, quadratic
pixcolor *= (1.0 - %(k_vignetting_strength)s*(0.5*sample_normsq));
"""

class Vignetting(Filter):
    """Filter that simulates vignetting (darkening of image periphery compared to the center).

    Introduced in Panda 1.9.0.

    """

    def __init__(self, **kwargs):
        super(Vignetting, self).__init__(**kwargs)

    def onReset(self):
        super(Vignetting, self).onReset()  # reset inherited properties

        self.sort = 70
        self.stageName   = "FilmOrDetector"
        self.isMergeable = True

        self.strength = 0.6

    def onAttachPipeline(self):
        self.registerInputTexture(texName="color")
        self.registerCustomInput(inputType="float",  inputName="k_vignetting_strength")

    def onSynthesizeCompositor(self):
        txcolor = self.getTextureInfo("color")
        code = VIGNETTING_BODY % { "k_vignetting_strength" : self.getMangledName("k_vignetting_strength"),
                                   "texcoord_txcolor"      : txcolor.texcoord,
                                   "texpad_txcolor"        : txcolor.texpad }
        return ("vignetting", code, "// vignetting filter\n")


    @property
    def strength(self):
        """Strength of the effect.

        Float, Default 0.6.

        For most cases, useful values are in the interval [0, 1].

        A value of 1.0 means the pixels at exactly the corners of the view will become completely black,
        while a value of 0.0 means no vignetting.

        Values over 1.0 make a larger part of the edge of the view black; e.g. 2.0 will make the pixels
        exactly at the halfway points of view edges completely black.

        """
        return self._strength
    @strength.setter
    def strength(self, value):
        self._strength = value
        if self.finalQuad is not None:
            self.finalQuad.setShaderInput( self.getMangledName("vignetting_strength"), value )

