from CompoundFilter import CompoundFilter
from Bloom import Bloom
from VolumetricLightingCompositor import VolumetricLightingCompositor


class VolumetricLighting(CompoundFilter):
    """A volumetric lighting ("god rays") filter.

    This filter is a compound, combining Bloom (to generate a glow texture) and VolumetricLightingCompositor
    (to compute and draw the rays based on the glow texture).

    This covers the most common use case, packaging it into a single all-in-one filter; for more exotic setups,
    it is still possible to use the combination of another filter and VolumetricLightingCompositor manually
    (or to create another compound filter to package that, following the example of this one).

    """

    def __init__(self, **kwargs):
        self.bloom = None
        self.volumetricLightingCompositor = None

        # We do *not* set self._customCompositing. This way CompoundFilter will use the last contained filter's
        # compositing logic to render this filter.

        # This calls createContainedFilters().
        super(VolumetricLighting, self).__init__(**kwargs)

    def onReset(self):
        super(VolumetricLighting, self).onReset()

        # Note that in a CompoundFilter, the isMergeable, stageName and sort of the contained filters
        # are simply ignored; the values of the host CompoundFilter are those that are used.
        #
        # We use the same default values for stageName and sort as VolumetricLightingCompositor does;
        # these filters are alternatives.
        #
        self.isMergeable = False
        self.stageName = "SceneOptics"
        self.sort = 60  # this should come after AmbientOcclusion, but can be done in the same render pass.

        # The idea of the VolumetricLighting compound filter is that it provides its own glow texture,
        # namely that generated by the contained Bloom filter. The value of "source" is fixed.
        #
        self.volumetricLightingCompositor.source = "bloomOutput"

        # Bloom defaults
        #
        self.bloomBlend = (0.3,0.4,0.3,0.0)
        self.mintrigger = 0.6
        self.maxtrigger = 1.0
        self.desat      = 0.6
        self.intensity  = 1.0
        self.bloomSize  = "medium"

        # VolumetricLightingCompositor defaults
        #
        self.caster     = None  # no sensible default possible
        self.numsamples = 32
        self.density    =  0.6
        self.decay      =  0.98
        self.exposure   =  0.1


    def createContainedFilters(self):
        # Here we create the contained filter instances and store them, in desired render order, in self.filters.
        #
        self.bloom = Bloom()
        self.volumetricLightingCompositor = VolumetricLightingCompositor()
        self.filters.append( self.bloom )
        self.filters.append( self.volumetricLightingCompositor )


    # For the properties, we pass through the implementation to the appropriate contained filter.

    # Bloom properties

    @property
    def bloomBlend(self):
        """[Glow texture bloom] The bloom filter needs to measure the brightness of each pixel. It does this by
        weighting the R,G,B and A components. Default weights: (0.3, 0.4, 0.3, 0.0).

        You should assign a nonzero weight to the alpha channel if you want the glow map
        to have an effect, or a value like (0, 0, 0, 1) if you only want your glow map
        to indicate which parts of a model should glow.

        """
        return self.bloom.blend
    @bloomBlend.setter
    def bloomBlend(self, value):
        self.bloom.blend = value

    @property
    def mintrigger(self):
        """[Glow texture bloom] Minimum brightness at which a halo is generated. Float, default 0.6."""
        return self.bloom.mintrigger
    @mintrigger.setter
    def mintrigger(self, value):
        self.bloom.mintrigger = value

    @property
    def maxtrigger(self):
        """[Glow texture bloom] Maximum brightness at which the halo reaches peak intensity. Float, default 1.0."""
        return self.bloom.maxtrigger
    @maxtrigger.setter
    def maxtrigger(self, value):
        self.bloom.maxtrigger = value

    @property
    def desat(self):
        """[Glow texture bloom] Degree to which the halo is desaturated. Float, in [0, 1]. Default 0.6.

        Setting this to zero means the halo is the same color as the bright pixel.
        Setting this to one means the halo is white.

        """
        return self.bloom.desat
    @desat.setter
    def desat(self, value):
        self.bloom.desat = value

    @property
    def intensity(self):
        """[Glow texture bloom] An adjustment parameter for the brightness of the halos. Float, default 1.0."""
        return self.bloom.intensity
    @intensity.setter
    def intensity(self, value):
        self.bloom.intensity = value

    @property
    def bloomSize(self):
        """[Glow texture bloom] Halo size. One of Bloom.SIZES ("small", "medium" [default] or "large").

        Controls the distance how far the halo effect reaches out from each blooming pixel.

        This is a discrete value instead of a continuous one, because the blur operation used
        internally in the bloom effect involves downsampling the original image by a power of two.

        """
        return self.bloom.size
    @bloomSize.setter
    def bloomSize(self, value):
        self.bloom.size = value


    # VolumetricLightingCompositor properties

    @property
    def caster(self):
        """NodePath that indicates the origin of the rays. Usually, you would pass your light,
        and create a sun billboard which is reparented to the light's NodePath.

        The billboard should be centered on the light's origin to correctly represent
        the light source. This is important because the rays shine outward from the
        origin of the caster NodePath (your light).

        There is no default value for this parameter.

        """
        return self.volumetricLightingCompositor.caster
    @caster.setter
    def caster(self, value):
        self.volumetricLightingCompositor.caster = value


    # We do not publish volumetricLightingCompositor.source, as that would defeat the purpose
    # of the contained Bloom filter.


    @property
    def numsamples(self):
        """Number of samples used to trace the light rays (default: 32).

        The more samples you use, the slower the effect will be, but you will have smoother light rays.
        Note that using a fuzzy billboarded dot instead of a hard-edged sphere as light caster can help
        with smoothing the end result. So does the use of the bloom filter as a preprocessor (see "source").

        This value does not need to be a power-of-two, it can be any positive number.

        """
        return self.volumetricLightingCompositor.numsamples
    @numsamples.setter
    def numsamples(self, value):
        self.volumetricLightingCompositor.numsamples = value


    @property
    def density(self):
        """Affects the length of the rays (default: 0.6).

        Usually a value between 0.5 and 1.0 works best. However, the best value for each particular case
        also depends on the values chosen for "numsamples" and "exposure".

        The default has changed in version 1.9.0; the backward-compatible CommonFilters API
        still uses the old default value of 5.0, which is much too high for most cases.

        """
        return self.volumetricLightingCompositor.density
    @density.setter
    def density(self, value):
        self.volumetricLightingCompositor.density = value


    @property
    def decay(self):
        """Affects how fast the rays decrease in brightness as they travel.

        Float, in interval (0,1). Smaller values make the rays decay faster (this acts as a damping coefficient).
        Usually, this should be a value close to 1.0, like 0.98 (default).

        The default has changed in version 1.9.0; the backward-compatible CommonFilters API
        still uses the old default value of 0.1, which is much too low for most cases.

        """
        return self.volumetricLightingCompositor.decay
    @decay.setter
    def decay(self, value):
        self.volumetricLightingCompositor.decay = value


    @property
    def exposure(self):
        """Defines the brightness of the rays (default 0.1)."""
        return self.volumetricLightingCompositor.exposure
    @exposure.setter
    def exposure(self, value):
        self.volumetricLightingCompositor.exposure = value

